#include <iostream>
#include <sstream>
#include <iomanip>
#include <cstdlib>
#include <assert.h>

#include <unistd.h>
#include <sys/time.h>

#include "CL/opencl.h"
#include "opencl.h"
#include "scoped_ptrs.h"

#include "bmp_util.h"

using namespace aocl_utils;

// ACL runtime configuration
static unsigned numDevices = 0;
static cl_platform_id platform;
static scoped_array<cl_device_id> devices;
static cl_context context;
static scoped_array<cl_command_queue> queues;
static scoped_array<cl_kernel> kernels;
static cl_program program;
static cl_int status;

static cl_mem theHard_sccb_dout;
static cl_mem theHard_image;

cl_event event;

int hardwareInitialize();
int hardwareRelease();
double gettimofday_sec();
int read_register(int addr);
void write_register(int addr, int din);
void set_all_registers();
void kick_capture();
void read_data(unsigned char *buf);

int main(int argc, char** argv){
  std::cout << "A simple sample." << std::endl;
 
  hardwareInitialize();
  const cl_event event_list[1] = {event};

  theHard_sccb_dout = clCreateBuffer(context, CL_MEM_WRITE_ONLY, sizeof(int), NULL, &status);
  checkError(status, "Failed to create input pixel buffer");
  theHard_image = clCreateBuffer(context, CL_MEM_WRITE_ONLY, 640*480*3*sizeof(unsigned char), NULL, &status);
  checkError(status, "Failed to create input pixel buffer");

  if(argc > 1){
    std::cout << "set registers." << std::endl;
    set_all_registers();
    std::cout << "...done" << std::endl;
  }

  unsigned char pixels[3*640*480];

  std::cout << "kick capture" << std::endl;
  kick_capture();
  std::cout << "...done" << std::endl;

  double t0 = gettimofday_sec();

  std::cout << "wait for all data" << std::endl;
  read_data(pixels);
  std::cout << "...done" << std::endl;

  double t1 = gettimofday_sec();

  write_bmp("test.bmp", pixels, 640, 480);

  double t2 = gettimofday_sec();

  clReleaseMemObject(theHard_sccb_dout);
  clReleaseMemObject(theHard_image);

  printf("capture: %f ms\n", (t1-t0)*1000);
  printf("write bmp: %f ms\n", (t2-t1)*1000);
  printf("total: %f ms\n", (t2-t0)*1000);

  hardwareRelease();

}

void kick_capture()
{
  status = clSetKernelArg(kernels[1], 0, sizeof(cl_mem), (void*)&theHard_image);
  checkError(status, "Failed to set kernel argument %d", 0);
  status = clEnqueueTask(queues[1], kernels[1], 0, NULL, &event);
  checkError(status, "Failed to enqueue task");

  int mode = 2; // kick capture
  int addr = 0;
  int din = 0;
  status = clSetKernelArg(kernels[0], 0, sizeof(int), (void*)&mode);
  checkError(status, "Failed to set kernel argument %d", 0);
  status = clSetKernelArg(kernels[0], 1, sizeof(int), (void*)&addr);
  checkError(status, "Failed to set kernel argument %d", 1);
  status = clSetKernelArg(kernels[0], 2, sizeof(int), (void*)&din);
  checkError(status, "Failed to set kernel argument %d", 2);
  status = clSetKernelArg(kernels[0], 3, sizeof(cl_mem), (void*)&theHard_sccb_dout);
  checkError(status, "Failed to set kernel argument %d", 3);

  status = clEnqueueTask(queues[0], kernels[0], 0, NULL, &event);
  checkError(status, "Failed to enqueue task");

}

void read_data(unsigned char *buf)
{

  status = clEnqueueBarrier(queues[1]);
  checkError(status, "Failed to enqueue barrier");
  status = clEnqueueReadBuffer(queues[1], theHard_image, CL_TRUE, 0, 3*640*480*sizeof(unsigned char), buf, 0, NULL, NULL);
  checkError(status, "Failed to read output");

  status = clEnqueueBarrier(queues[0]);
  checkError(status, "Failed to enqueue barrier");
  status = clEnqueueBarrier(queues[1]);
  checkError(status, "Failed to enqueue barrier");

  const cl_event event_list[1] = {event};
  clWaitForEvents(1, event_list);
}

int read_register(int addr)
{
  int mode = 0;
  int din = 0;
  int dout = 0;
  status = clSetKernelArg(kernels[0], 0, sizeof(int), (void*)&mode);
  checkError(status, "Failed to set kernel argument %d", 0);
  status = clSetKernelArg(kernels[0], 1, sizeof(int), (void*)&addr);
  checkError(status, "Failed to set kernel argument %d", 1);
  status = clSetKernelArg(kernels[0], 2, sizeof(int), (void*)&din);
  checkError(status, "Failed to set kernel argument %d", 2);
  status = clSetKernelArg(kernels[0], 3, sizeof(cl_mem), (void*)&theHard_sccb_dout);
  checkError(status, "Failed to set kernel argument %d", 3);

  //std::cout << "read_register: setKernelArg...done" << std::endl;

  status = clEnqueueTask(queues[0], kernels[0], 0, NULL, &event);
  checkError(status, "Failed to enqueue task");
  //std::cout << "read_register: enqueueTask...done" << std::endl;

  status = clEnqueueBarrier(queues[0]);
  checkError(status, "Failed to enqueue barrier");
  //std::cout << "read_register: enqueueBarrier...done" << std::endl;
  status = clEnqueueReadBuffer(queues[0], theHard_sccb_dout, CL_TRUE, 0, sizeof(int), &dout, 0, NULL, NULL);
  checkError(status, "Failed to read output");
  //std::cout << "read_register: enqueueReadBuffer...done" << std::endl;

  status = clEnqueueBarrier(queues[0]);
  checkError(status, "Failed to enqueue barrier");

  const cl_event event_list[1] = {event};
  clWaitForEvents(1, event_list);

  return dout;
}

void write_register(int addr, int din)
{
  int mode = 1; // WRITE_MODE
  int dout = 0;
  status = clSetKernelArg(kernels[0], 0, sizeof(int), (void*)&mode);
  checkError(status, "Failed to set kernel argument %d", 0);
  status = clSetKernelArg(kernels[0], 1, sizeof(int), (void*)&addr);
  checkError(status, "Failed to set kernel argument %d", 1);
  status = clSetKernelArg(kernels[0], 2, sizeof(int), (void*)&din);
  checkError(status, "Failed to set kernel argument %d", 2);
  status = clSetKernelArg(kernels[0], 3, sizeof(cl_mem), (void*)&theHard_sccb_dout);
  checkError(status, "Failed to set kernel argument %d", 3);
  //std::cout << "write_register: setKernelArg...done" << std::endl;

  status = clEnqueueTask(queues[0], kernels[0], 0, NULL, &event);
  checkError(status, "Failed to enqueue task");
  //std::cout << "write_register: enqueueTask...done" << std::endl;

  status = clEnqueueBarrier(queues[0]);
  checkError(status, "Failed to enqueue barrier");
  //std::cout << "write_register: enqueueBarrier...done" << std::endl;
  status = clEnqueueReadBuffer(queues[0], theHard_sccb_dout, CL_TRUE, 0, sizeof(int), &dout, 0, NULL, NULL);
  checkError(status, "Failed to read output");
  //std::cout << "write_register: enqueueReadBuffer...done" << std::endl;

  status = clEnqueueBarrier(queues[0]);
  checkError(status, "Failed to enqueue barrier");

  const cl_event event_list[1] = {event};
  clWaitForEvents(1, event_list);
}

double gettimofday_sec(void){
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return ((double)(tv.tv_sec) + (double)(tv.tv_usec) * 0.001 * 0.001);
}

int hardwareInitialize()
{
  if(!setCwdToExeDir()) 
  {
    return -1;
  }

  // Set up the platform
  platform = findAnyPlatform();
  if(platform == NULL)
  {
    printf("Found no platforms!\n");
    hardwareRelease();
    return -1;
  }

  // Set up the device(s)
  devices.reset(getDevices(platform, CL_DEVICE_TYPE_ALL, &numDevices));

  // Print the name of the platform being used
  printf("Using platform: %s\n", getPlatformName(platform).c_str());
  printf("Using %d devices:\n", numDevices);
  for(unsigned i = 0; i < numDevices; ++i) {
    printf("  %s\n", getDeviceName(devices[i]).c_str());
  }
  
  // Create a context
  context = clCreateContext(0, numDevices, devices, &oclContextCallback, NULL, &status);
  checkError(status, "Failed to create context");

  // Create command queues
  queues.reset((uint)0);
  queues[0] = clCreateCommandQueue(context, devices[0], CL_QUEUE_PROFILING_ENABLE, &status);
  checkError(status, "Failed to create command queue");
  queues[1] = clCreateCommandQueue(context, devices[0], CL_QUEUE_PROFILING_ENABLE, &status);
  checkError(status, "Failed to create command queue");

  if(isAlteraPlatform(platform))
  {
    //Create the program using the binary aocx file
    printf("Create the program using the binary aocx file\n");
    std::string binary_file = getBoardBinaryFile("simple_sample", devices[0]);
    program = createProgramFromBinary(context, binary_file.c_str(), devices, numDevices);
  }
  else
  {
    std::string binary_file = "../device/simple_sample.cl";
    printf("Using AOCX: %s\n", binary_file.c_str());
    program = createProgramFromSource(context, binary_file.c_str(), devices, numDevices, "-DDONT_USE_PRAGMA");
  }

  // Create the kernels
  kernels.reset((uint)0);

  kernels[0] = clCreateKernel(program, "ov7670_ctrl", &status);
  checkError(status, "Failed to create kernel");
  kernels[1] = clCreateKernel(program, "ov7670_capture", &status);
  checkError(status, "Failed to create kernel");

  // Return success
  return 0;
}

// free memory allocated by the program
int hardwareRelease()
{
  // Release all created objects
  if(kernels && kernels[0]) clReleaseKernel(kernels[0]);
  if(kernels && kernels[1]) clReleaseKernel(kernels[1]);

  if(queues && queues[0]) clReleaseCommandQueue(queues[0]);
  if(queues && queues[1]) clReleaseCommandQueue(queues[1]);

  if(program) 
    clReleaseProgram(program);
  if(context) 
    clReleaseContext(context);

  // Return success
  return 0;
}

// Called by aocl_utils::checkError
void cleanup() {
  hardwareRelease();
}


#include "registers.h"

void set_register(int addr, int value, bool preserve_flag)
{
  int v;
  if(preserve_flag){
    v = read_register(addr);
    v = (v & RESERVED_PATTERN[addr]) | (value & 0x000000FF);
  }else{
    v = value & 0x000000FF;
  }
  write_register(addr, v);

  int result = read_register(addr);
  if(v != result){
    printf("@%02x: %02x => %02x <> %02x\n", addr, value, v, result);
    if(!(addr == 0x12 || addr == 0xc9)){
      set_register(addr, value, preserve_flag);
    }
  }
}

void set_all_registers(){
  set_register(REG_COM7, COM7_RESET, false);
  usleep(100000); // 1ms待つ (ここでは，100m秒くらい待っている)

  printf("default_register_values\n");
  int n = sizeof(default_register_values) / sizeof(default_register_values[0]);
  for(int i = 0; i < n; i++){
    set_register(default_register_values[i][0], default_register_values[i][1], false);
  }

  printf("yuv422_values\n");
  n = sizeof(yuv422_values) / sizeof(yuv422_values[0]);
  for(int i = 0; i < n; i++){
    set_register(yuv422_values[i][0], yuv422_values[i][1], true);
  }

  printf("color_matrix_yuv422\n");
  n = sizeof(color_matrix_yuv422) / sizeof(color_matrix_yuv422);
  for(int i = 0; i < n; i++){
    set_register(color_matrix_yuv422[i][0], color_matrix_yuv422[i][1], true);
  }
}
